use std::collections::{hash_map::Entry,HashMap};

use mlua::{FromLua,FromLuaMulti,IntoLua,IntoLuaMulti};
use rbx_types::Ref;
use rbx_dom_weak::{InstanceBuilder,WeakDom};

use super::vector3::Vector3;

pub fn set_globals(lua:&mlua::Lua,globals:&mlua::Table)->Result<(),mlua::Error>{
	//class functions store
	lua.set_app_data(ClassFunctions::default());

	let instance_table=lua.create_table()?;

	//Instance.new
	instance_table.raw_set("new",
		lua.create_function(|lua,(class_name,parent):(mlua::String,Option<Instance>)|{
			let class_name_str=&*class_name.to_str()?;
			let parent=parent.ok_or(mlua::Error::runtime("Nil Parent not yet supported"))?;
			dom_mut(lua,|dom|{
				//TODO: Nil instances
				Ok(Instance::new(dom.insert(parent.referent,InstanceBuilder::new(class_name_str))))
			})
		})?
	)?;

	globals.set("Instance",instance_table)?;

	Ok(())
}

// LMAO look at this function!
fn dom_mut<T>(lua:&mlua::Lua,mut f:impl FnMut(&mut WeakDom)->mlua::Result<T>)->mlua::Result<T>{
	let mut dom=lua.app_data_mut::<&'static mut WeakDom>().ok_or(mlua::Error::runtime("DataModel missing"))?;
	f(&mut *dom)
}
fn class_functions_mut<T>(lua:&mlua::Lua,mut f:impl FnMut(&mut ClassFunctions)->mlua::Result<T>)->mlua::Result<T>{
	let mut cf=lua.app_data_mut::<ClassFunctions>().ok_or(mlua::Error::runtime("ClassFunctions missing"))?;
	f(&mut *cf)
}

fn coerce_float32(value:&mlua::Value)->Option<f32>{
	match value{
		&mlua::Value::Integer(i)=>Some(i as f32),
		&mlua::Value::Number(f)=>Some(f as f32),
		_=>None,
	}
}

fn get_full_name(dom:&rbx_dom_weak::WeakDom,instance:&rbx_dom_weak::Instance)->String{
	let mut full_name=instance.name.clone();
	let mut pref=instance.parent();
	while let Some(parent)=dom.get_by_ref(pref){
		full_name.insert(0,'.');
		full_name.insert_str(0,parent.name.as_str());
		pref=parent.parent();
	}
	full_name
}
//helper function for script
pub fn get_name_source(lua:&mlua::Lua,script:Instance)->Result<(String,String),mlua::Error>{
	dom_mut(lua,|dom|{
		let instance=script.get(dom)?;
		let source=match instance.properties.get("Source"){
			Some(rbx_dom_weak::types::Variant::String(s))=>s.clone(),
			_=>Err(mlua::Error::external("Missing script.Source"))?,
		};
		Ok((get_full_name(dom,instance),source))
	})
}

pub fn find_first_child<'a>(dom:&'a rbx_dom_weak::WeakDom,instance:&rbx_dom_weak::Instance,name:&str)->Option<&'a rbx_dom_weak::Instance>{
	instance.children().iter().filter_map(|&r|dom.get_by_ref(r)).find(|inst|inst.name==name)
}
pub fn find_first_descendant<'a>(dom:&'a rbx_dom_weak::WeakDom,instance:&rbx_dom_weak::Instance,name:&str)->Option<&'a rbx_dom_weak::Instance>{
	dom.descendants_of(instance.referent()).find(|&inst|inst.name==name)
}

pub fn find_first_child_of_class<'a>(dom:&'a rbx_dom_weak::WeakDom,instance:&rbx_dom_weak::Instance,class:&str)->Option<&'a rbx_dom_weak::Instance>{
	instance.children().iter().filter_map(|&r|dom.get_by_ref(r)).find(|inst|inst.class==class)
}
pub fn find_first_descendant_of_class<'a>(dom:&'a rbx_dom_weak::WeakDom,instance:&rbx_dom_weak::Instance,class:&str)->Option<&'a rbx_dom_weak::Instance>{
	dom.descendants_of(instance.referent()).find(|&inst|inst.class==class)
}

#[derive(Clone,Copy)]
pub struct Instance{
	referent:Ref,
}
impl Instance{
	pub const fn new(referent:Ref)->Self{
		Self{referent}
	}
	pub fn get<'a>(&self,dom:&'a WeakDom)->mlua::Result<&'a rbx_dom_weak::Instance>{
		dom.get_by_ref(self.referent).ok_or(mlua::Error::runtime("Instance missing"))
	}
	pub fn get_mut<'a>(&self,dom:&'a mut WeakDom)->mlua::Result<&'a mut rbx_dom_weak::Instance>{
		dom.get_by_ref_mut(self.referent).ok_or(mlua::Error::runtime("Instance missing"))
	}
}
type_from_lua_userdata!(Instance);

//TODO: update rbx_reflection and use dom.superclasses_iter
pub struct SuperClassIter<'a> {
    database: &'a rbx_reflection::ReflectionDatabase<'a>,
    descriptor: Option<&'a rbx_reflection::ClassDescriptor<'a>>,
}
impl<'a> SuperClassIter<'a> {
    fn next_descriptor(&self) -> Option<&'a rbx_reflection::ClassDescriptor<'a>> {
        let superclass = self.descriptor?.superclass.as_ref()?;
        self.database.classes.get(superclass)
    }
}
impl<'a> Iterator for SuperClassIter<'a> {
    type Item = &'a rbx_reflection::ClassDescriptor<'a>;
    fn next(&mut self) -> Option<Self::Item> {
        let next_descriptor = self.next_descriptor();
        std::mem::replace(&mut self.descriptor, next_descriptor)
    }
}

impl mlua::UserData for Instance{
	fn add_fields<F:mlua::UserDataFields<Self>>(fields:&mut F){
		fields.add_field_method_get("Parent",|lua,this|{
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(Instance::new(instance.parent()))
			})
		});
		fields.add_field_method_set("Parent",|lua,this,val:Option<Instance>|{
			let parent=val.ok_or(mlua::Error::runtime("Nil Parent not yet supported"))?;
			dom_mut(lua,|dom|{
				dom.transfer_within(this.referent,parent.referent);
				Ok(())
			})
		});
		fields.add_field_method_get("Name",|lua,this|{
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(instance.name.clone())
			})
		});
		fields.add_field_method_set("Name",|lua,this,val:mlua::String|{
			dom_mut(lua,|dom|{
				let instance=this.get_mut(dom)?;
				//Why does this need to be cloned?
				instance.name=val.to_str()?.to_owned();
				Ok(())
			})
		});
		fields.add_field_method_get("ClassName",|lua,this|{
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(instance.class.clone())
			})
		});
	}
	fn add_methods<M:mlua::UserDataMethods<Self>>(methods:&mut M){
		methods.add_method("GetChildren",|lua,this,_:()|
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				let children:Vec<_>=instance
					.children()
					.iter()
					.copied()
					.map(Instance::new)
					.collect();
				Ok(children)
			})
		);
		fn ffc(lua:&mlua::Lua,this:&Instance,(name,search_descendants):(mlua::String,Option<bool>))->mlua::Result<Option<Instance>>{
			let name_str=&*name.to_str()?;
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(
					match search_descendants.unwrap_or(false){
						true=>find_first_descendant(dom,instance,name_str),
						false=>find_first_child(dom,instance,name_str),
					}
					.map(|instance|
						Instance::new(instance.referent())
					)
				)
			})
		}
		methods.add_method("FindFirstChild",ffc);
		methods.add_method("WaitForChild",ffc);
		methods.add_method("FindFirstChildOfClass",|lua,this,(class,search_descendants):(mlua::String,Option<bool>)|{
			let class_str=&*class.to_str()?;
			dom_mut(lua,|dom|{
				Ok(
					match search_descendants.unwrap_or(false){
						true=>find_first_descendant_of_class(dom,this.get(dom)?,class_str),
						false=>find_first_child_of_class(dom,this.get(dom)?,class_str),
					}
					.map(|instance|
						Instance::new(instance.referent())
					)
				)
			})
		});
		methods.add_method("GetDescendants",|lua,this,_:()|
			dom_mut(lua,|dom|{
				let children:Vec<_>=dom
					.descendants_of(this.referent)
					.map(|instance|
						Instance::new(instance.referent())
					)
					.collect();
				Ok(children)
			})
		);
		methods.add_method("IsA",|lua,this,classname:mlua::String|
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(crate::context::class_is_a(instance.class.as_str(),&*classname.to_str()?))
			})
		);
		methods.add_method("Destroy",|lua,this,()|
			dom_mut(lua,|dom|{
				dom.destroy(this.referent);
				Ok(())
			})
		);
		methods.add_meta_function(mlua::MetaMethod::ToString,|lua,this:Instance|{
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				Ok(instance.name.clone())
			})
		});
		methods.add_meta_function(mlua::MetaMethod::Index,|lua,(this,index):(Instance,mlua::String)|{
			let index_str=&*index.to_str()?;
			dom_mut(lua,|dom|{
				let instance=this.get(dom)?;
				//println!("__index t={} i={index:?}",instance.name);
				let db=rbx_reflection_database::get();
				let class=db.classes.get(instance.class.as_str()).ok_or(mlua::Error::runtime("Class missing"))?;
				//Find existing property
				match instance.properties.get(index_str)
					.cloned()
					//Find default value
					.or_else(||db.find_default_property(class,index_str).cloned())
					//Find virtual property
					.or_else(||{
						SuperClassIter{
							database:db,
							descriptor:Some(class),
						}
						.find_map(|class|
							find_virtual_property(&instance.properties,class,index_str)
						)
					})
				{
					Some(rbx_types::Variant::Int32(val))=>return val.into_lua(lua),
					Some(rbx_types::Variant::Int64(val))=>return val.into_lua(lua),
					Some(rbx_types::Variant::Float32(val))=>return val.into_lua(lua),
					Some(rbx_types::Variant::Float64(val))=>return val.into_lua(lua),
					Some(rbx_types::Variant::Ref(val))=>return Instance::new(val).into_lua(lua),
					Some(rbx_types::Variant::CFrame(cf))=>return Into::<super::cframe::CFrame>::into(cf).into_lua(lua),
					Some(rbx_types::Variant::Vector3(v))=>return Into::<super::vector3::Vector3>::into(v).into_lua(lua),
					None=>(),
					other=>return Err(mlua::Error::runtime(format!("Instance.__index Unsupported property type instance={} index={index_str} value={other:?}",instance.name))),
				}
				//find a function with a matching name
				if let Some(function)=class_functions_mut(lua,|cf|{
					let mut iter=SuperClassIter{
						database:db,
						descriptor:Some(class),
					};
					iter.find_map(|class|{
						let mut class_methods=cf.get_or_create_class_methods(&class.name)?;
						class_methods.get_or_create_function(lua,index_str)
							.transpose()
					}).transpose()
				})?{
					return function.into_lua(lua);
				}
				//find a child with a matching name
				find_first_child(dom,instance,index_str)
				.map(|instance|Instance::new(instance.referent()))
				.into_lua(lua)
			})
		});
		methods.add_meta_function(mlua::MetaMethod::NewIndex,|lua,(this,index,value):(Instance,mlua::String,mlua::Value)|{
			dom_mut(lua,|dom|{
				let instance=this.get_mut(dom)?;
				//println!("__newindex t={} i={index:?} v={value:?}",instance.name);
				let index_str=&*index.to_str()?;
				let db=rbx_reflection_database::get();
				let class=db.classes.get(instance.class.as_str()).ok_or(mlua::Error::runtime("Class missing"))?;
				let mut iter=SuperClassIter{
					database:db,
					descriptor:Some(class),
				};
				let property=iter.find_map(|cls|cls.properties.get(index_str)).ok_or(mlua::Error::runtime(format!("Property '{index_str}' missing on class '{}'",class.name)))?;
				match &property.data_type{
					rbx_reflection::DataType::Value(rbx_types::VariantType::Vector3)=>{
						let typed_value:Vector3=*value.as_userdata().ok_or(mlua::Error::runtime("Expected Userdata"))?.borrow()?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::Vector3(typed_value.into()));
					},
					rbx_reflection::DataType::Value(rbx_types::VariantType::Float32)=>{
						let typed_value:f32=coerce_float32(&value).ok_or(mlua::Error::runtime("Expected f32"))?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::Float32(typed_value));
					},
					rbx_reflection::DataType::Enum(enum_name)=>{
						let typed_value=match &value{
							&mlua::Value::Integer(int)=>Ok(rbx_types::Enum::from_u32(int as u32)),
							&mlua::Value::Number(num)=>Ok(rbx_types::Enum::from_u32(num as u32)),
							mlua::Value::String(s)=>{
								let e=db.enums.get(enum_name).ok_or(mlua::Error::runtime("Database DataType Enum name  does not exist"))?;
								Ok(rbx_types::Enum::from_u32(*e.items.get(&*s.to_str()?).ok_or(mlua::Error::runtime("Invalid enum item"))?))
							},
							mlua::Value::UserData(any_user_data)=>{
								let e:super::r#enum::Enum=*any_user_data.borrow()?;
								Ok(e.into())
							},
							_=>Err(mlua::Error::runtime("Expected Enum")),
						}?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::Enum(typed_value));
					},
					rbx_reflection::DataType::Value(rbx_types::VariantType::Color3)=>{
						let typed_value:super::color3::Color3=*value.as_userdata().ok_or(mlua::Error::runtime("Expected Color3"))?.borrow()?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::Color3(typed_value.into()));
					},
					rbx_reflection::DataType::Value(rbx_types::VariantType::Bool)=>{
						let typed_value=value.as_boolean().ok_or(mlua::Error::runtime("Expected boolean"))?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::Bool(typed_value));
					},
					rbx_reflection::DataType::Value(rbx_types::VariantType::String)=>{
						let typed_value=value.as_str().ok_or(mlua::Error::runtime("Expected boolean"))?;
						instance.properties.insert(index_str.to_owned(),rbx_types::Variant::String(typed_value.to_owned()));
					},
					other=>return Err(mlua::Error::runtime(format!("Unimplemented property type: {other:?}"))),
				}
				Ok(())
			})
		});
	}
}

/// A class function definition shorthand.
macro_rules! cf{
	($f:expr)=>{
		|lua,mut args|{
			let this=Instance::from_lua(args.pop_front().unwrap_or(mlua::Value::Nil),lua)?;
			$f(lua,this,FromLuaMulti::from_lua_multi(args,lua)?)?.into_lua_multi(lua)
		}
	};
}
type ClassFunctionPointer=fn(&mlua::Lua,mlua::MultiValue)->mlua::Result<mlua::MultiValue>;
// TODO: use macros to define these with better organization
/// A double hash map of function pointers.
/// The class tree is walked by the Instance.__index metamethod to find available class methods.
type CFD=phf::Map<&'static str,// Class name
	phf::Map<&'static str,// Method name
		ClassFunctionPointer
	>
>;
static CLASS_FUNCTION_DATABASE:CFD=phf::phf_map!{
	"DataModel"=>phf::phf_map!{
		"GetService"=>cf!(|lua,_this,service:mlua::String|{
			dom_mut(lua,|dom|{
				//dom.root_ref()==this.referent ?
				match &*service.to_str()?{
					"Lighting"=>{
						let referent=find_first_child_of_class(dom,dom.root(),"Lighting")
        				.map(|instance|instance.referent())
						.unwrap_or_else(||
							dom.insert(dom.root_ref(),InstanceBuilder::new("Lighting"))
						);
						Ok(Instance::new(referent))
					},
					other=>Err::<Instance,_>(mlua::Error::runtime(format!("Service '{other}' not supported"))),
				}
			})
		}),
	},
	"Terrain"=>phf::phf_map!{
		"FillBlock"=>cf!(|_lua,_,_:(super::cframe::CFrame,Vector3,super::r#enum::Enum)|mlua::Result::Ok(()))
	},
};

/// A store of created functions for each Roblox class.
/// Functions are created the first time they are accessed and stored in this data structure.
#[derive(Default)]
struct ClassFunctions{
	classes:HashMap<&'static str,//ClassName
		HashMap<&'static str,//Method name
			mlua::Function
		>
	>
}
impl ClassFunctions{
	/// return self.classes[class] or create the ClassMethods and then return it
	fn get_or_create_class_methods(&mut self,class:&str)->Option<ClassMethods>{
		// Use get_entry to get the &'static str keys of the database
		// and use it as a key for the classes hashmap
		CLASS_FUNCTION_DATABASE.get_entry(class)
			.map(|(&static_class_str,method_pointers)|
				ClassMethods{
					method_pointers,
					methods:self.classes.entry(static_class_str)
					.or_insert_with(||HashMap::new()),
				}
			)
	}
}
struct ClassMethods<'a>{
	method_pointers:&'static phf::Map<&'static str,ClassFunctionPointer>,
	methods:&'a mut HashMap<&'static str,mlua::Function>,
}
impl ClassMethods<'_>{
	/// return self.methods[index] or create the function in the hashmap and then return it
	fn get_or_create_function(&mut self,lua:&mlua::Lua,index:&str)->mlua::Result<Option<mlua::Function>>{
		Ok(match self.method_pointers.get_entry(index){
			Some((&static_index_str,&function_pointer))=>Some(
				match self.methods.entry(static_index_str){
					Entry::Occupied(entry)=>entry.get().clone(),
					Entry::Vacant(entry)=>entry.insert(
						lua.create_function(function_pointer)?
					).clone(),
				}
			),
			None=>None,
		})
	}
}

/// A virtual property pointer definition shorthand.
type VirtualPropertyFunctionPointer=fn(&rbx_types::Variant)->Option<rbx_types::Variant>;
const fn vpp(
	property:&'static str,
	pointer:VirtualPropertyFunctionPointer,
)->VirtualProperty{
	VirtualProperty{
		property,
		pointer,
	}
}
struct VirtualProperty{
	property:&'static str,// Source property name
	pointer:VirtualPropertyFunctionPointer,
}
type VPD=phf::Map<&'static str,// Class name
	phf::Map<&'static str,// Virtual property name
		VirtualProperty
	>
>;
static VIRTUAL_PROPERTY_DATABASE:VPD=phf::phf_map!{
	"BasePart"=>phf::phf_map!{
		"Position"=>vpp("CFrame",|c:&rbx_types::Variant|{
			let c=match c{
				rbx_types::Variant::CFrame(c)=>c,
				_=>return None,//fail silently and ungracefully
			};
			Some(rbx_types::Variant::Vector3(c.position))
		}),
	},
};

fn find_virtual_property(
	properties:&HashMap<String,rbx_types::Variant>,
	class:&rbx_reflection::ClassDescriptor,
	index:&str
)->Option<rbx_types::Variant>{
	//Find virtual property
	let class_virtual_properties=VIRTUAL_PROPERTY_DATABASE.get(&class.name)?;
	let virtual_property=class_virtual_properties.get(index)?;

	//Get source property
	let variant=properties.get(virtual_property.property)?;

	//Transform Source property with provided function
	(virtual_property.pointer)(variant)
}
